# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007  
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr 
#  
# dataproxy.py
#  
# classes permettant de vérifier l'authentification sur différents supports
# (actuellement seulement annuaire LDAP)
#       
###########################################################################

# utile ?
class IAuthProxy:
    """Interface pour un système d'authentification et de récupération des infos utilisateurs.
    Mettre les éventuels commandes de connextion, ... dans __init__.py"""
       
    def get_data(self, identifier):
        """Récupère les données correspondant à identifier
        
        @param identifier: à priori l'identifiant d'un utilisateur dans le système de données
        """
        pass

    def authenticate(self, login, password):
        """Vérifie la validité d'un couple login/password

        @param login: identifiant de l'utilisateur dans ce système
        @param password: mot de passe

        retourne True ou False 
        """
        pass
        
import ldap

class LDAPProxy:
    """récupère les infos d'un utilisateur dans un annuaire LDAP"""
    __implements__ = (IAuthProxy,)

    ignored_attrs = ['userPassword']

    def __init__(self, hosts, ports, bases, static_data={}):
        """
        hosts : liste de serveurs ldap (ou un seul)
        ports : liste des ports correspondants
        """
        self.ldap_servers = []
        if type(hosts) is not list:
            hosts = [hosts]
        if type(ports) is not list:
            ports = [ports]
        if type(bases) is not list:
            bases = [bases]
        for host, port, base in zip(hosts, ports, bases):
            self.ldap_servers.append((host, port, base))
        # on attend des listes de valeur pour hosts  et ports
        # si ancienne configuration, on met les valeurs uniques dans une liste
        self.static_data = static_data
        self.proxy = ldap.open(hosts[0], int(ports[0]))
            
    def _proxy_reload(self, host, port):
        """
        recharge la connexion au ldap
        """
        self.proxy.unbind()
        self.proxy = ldap.open(host, int(port))

    def simple_bind(self, search_result, passwd):
        """
        essaye un bind sur le proxy de travail
        """
        try:
            user_dn = search_result[0][0]
            self.proxy.simple_bind_s(user_dn, passwd)
        except Exception, e:
            success = False
        else:
            success = True
        return success
            
    def authenticate(self, uid, passwd):
        # gestion multi ldap
        for host, port, base in self.ldap_servers:
            self._proxy_reload(host, port)
            ret = {}
            result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE,'(uid=%s)' % uid)
            success = self.simple_bind(result, passwd)
            # si l'utilisateur n'est pas authentifié, on essaie le serveur ldap suivant
            if success:
                break

        if success:
            try:
                ## renvoi des groupes de l'utilisateur en minuscule
                user_groups = []
                group_filters = "(&(objectclass=sambaGroupMapping)(objectclass=posixGroup)(memberUid=%s))"%uid
                group_result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE, group_filters, ['cn'])
                for gr in group_result:
                    user_groups.append(gr[1]['cn'][0].lower())
            except:
                user_groups = []
        # récupération des données 
        try:
            ret = result[0][1]
            # on supprime certains champs (mot de passe)
            for user_attr in self.ignored_attrs:
                if user_attr in ret:
                    del(ret[user_attr])
            ret['user_groups'] = user_groups
            ret.update(self.static_data)
        except:
            ret = {}
        # self.proxy.unbind_s()
        return success, ret

    #def get_data(self, uid):
    #    result = self.proxy.search_s(self.base, ldap.SCOPE_SUBTREE, '(uid=%s)' % uid)
    #    try:
    #        ## renvoi des groupes de l'utilisateur en minuscule
    #        user_groups = []
    #        group_filters = "(&(objectclass=sambaGroupMapping)(objectclass=posixGroup)(memberUid=%s))"%uid
    #        group_result = self.proxy.search_s(self.base, ldap.SCOPE_SUBTREE, group_filters, ['cn'])
    #        for gr in group_result:
    #            user_groups.append(gr[1]['cn'][0].lower())
    #    except:
    #        user_groups = []
    #    # récupération des données 
    #    # FIXME : utiliser le compte utilisateur ou un compte anonyme ?
    #    try:
    #        ret = result[0][1]
    #        # on supprime certains champs (mot de passe)
    #        for user_attr in self.ignored_attrs:
    #            if user_attr in ret:
    #                del(ret[user_attr])
    #        ret['user_groups'] = user_groups
    #        ret.update(self.static_data)
    #        return ret
    #    except:
    #        return {}
