#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# saml_idp.py
#
# support (limité) du protocole SAML 2.0 pour eole-sso
#
#
###########################################################################

import traceback
from xmlrpclib import escape
import urllib, datetime, base64

from config import CERTFILE, AUTH_FORM_URL, AUTH_SERVER_ADDR, IDP_IDENTITY, encoding, DEFAULT_CSS
from saml_utils import date_from_string
from saml_crypto import sign_request, check_signed_request, check_signed_doc
import saml_response
from saml_utils import encode_request, decode_request, gen_metadata, get_endpoint, InternalError, Redirect, check_date
from saml2 import saml, samlp
from eolesso.util import getCookies, gen_random_id, gen_ticket_id, gen_page, format_err, get_service_from_url

from twisted.python import log
from twisted.web2 import http, responsecode
from twisted.web2.http_headers import MimeType
from twisted.web2.resource import PostableResource

debug = 1

#try:
#    _ = _
#except NameError:
#    _ = unicode

class SamlResource(PostableResource):
    content_type = 'text/html'

    def render(self, request):
        headers = {'Content-type': MimeType.fromString(self.content_type)}
        errheaders = {'Content-type': MimeType.fromString('application/xml')}
        try:
            return self._render(request)
        except Redirect, exc:
            location = exc.location
            headers['location'] =  location
            log.msg('redirecting to %s', location)
            # 303 See other
            return http.Response(code=303, headers=headers)
        except Exception, exc: # unknown error
            traceback.print_exc()
            return http.StatusResponse(code=http.responsecode.INTERNAL_SERVER_ERROR, description=str(exc))
            # return http.Response(stream=error.as_xml(), headers=errheaders)

def extract_message(CERTFILE, request, message_type):
    # FIXME : utiliser un certificat partagé avec le service provider
    if 'Signature' in request.args:
        code, reason = check_signed_request(CERTFILE, request.args, message_type)
        if code or debug:
            xml_doc = decode_request(request.args[message_type][0], True)
            return True, xml_doc
    else:
        xml_doc = decode_request(request.args[message_type][0])
        code, reason = check_signed_doc(CERTFILE, xml_doc)
        if code or debug:
            return True, xml_doc
    return False, reason

class SamlLogout(SamlResource):
    """Resource de gestion du logout (logout cas + single logout saml)
    """
    addSlash = False

    def __init__(self, manager):
        self.manager = manager
        super(SamlLogout, self).__init__()

    def send_logout_request(self, sp_identity, sp_meta, user_id, ticket, sso_session, from_url):
        """sends a logout request to a service provider and store ticket and original request"""
        try:
            service_url, binding, response_url = get_endpoint(sp_meta, 'SingleLogoutService')
        except Exception, e:
            # store error for display ?
            return False

        response_id = gen_random_id('_')
        session_index = ticket.ticket
        from_string = ""
        if from_url:
            from_string = "%s" % urllib.urlencode([('url',from_url)])
            #from_string = "?%s" % urllib.urlencode([('url',from_url)])
        # sending request to corresponding entity
        if binding == samlp.BINDING_HTTP_POST:
            sign_method, req = saml_response.gen_logout_request(response_id, session_index, user_id, IDP_IDENTITY, sp_meta['entityID'], service_url, CERTFILE)
            # POST response
            req = encode_request(req)
            form="""
                    <FORM action="%s" method="POST">
                    <input type="hidden" name="SAMLRequest" value="%s">
                    <input type="hidden" name="SigAlg" value="%s">
                    <input type="hidden" name="RelayState" value="%s">
                    <input type="Submit" value="%s">
            """ % (service_url, req, sign_method, from_url, _("Proceed to service"))
            #""" % (service_url, req, sign_method, '%s/logout%s' % (AUTH_FORM_URL, from_string), _("Proceed to service"))
            resp = http.Response(stream=gen_page(_('Logout Request to service'), form, css))
        else:
            sign_method, req = saml_response.gen_logout_request(response_id, session_index, user_id, IDP_IDENTITY, sp_meta['entityID'], service_url, CERTFILE, sign=False)
            # REDIRECT response
            req = encode_request(req, True)
            req_args = sign_request(CERTFILE, sign_method, 'SAMLRequest', req, from_url)
            #req_args = sign_request(CERTFILE, sign_method, 'SAMLRequest', req, '%s/logout%s' % (AUTH_FORM_URL, from_string))
            if req_args != []:
                redirect_url = '%s?%s' % (service_url, urllib.urlencode(req_args))
            else:
                redirect_url = service_url
            resp = http.RedirectResponse(redirect_url)

        # before redirection, we store request id and ticket
        ticket.valid = False
        self.manager.pending_logout[sso_session][3][response_id] = ticket.ticket
        log.msg("sending logout request to %s for %s" % (sp_identity, ticket.service_url))
        return resp

    def _render(self, request):
        """SSO Logout response"""
        request_date = datetime.datetime.utcnow()
        # we check that user is logged_in
        authenticated = False
        cookies = getCookies(request)
        sso_session = None
        err_msg = None
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                sso_session = cookie.value
                # we check this session validity
                authenticated = self.manager.validate_session(sso_session)
                break

        try:
            css = request.args['css'][0]
        except KeyError:
            css = DEFAULT_CSS
        # check if we have to redirect after logout
        from_url = None
        relay_state = None
        if 'RelayState' in request.args:
            relay_state = request.args['RelayState'][0]
            from_url = relay_state
        elif request.args.has_key('service'):
            from_url = request.args['service'][0]
        elif request.args.has_key('url'):
            from_url = request.args['url'][0]
        closed_services = []
        failed_services = []

        try:
            ########################
            # Gestion du logout SAML
            if not authenticated:
                # pas d'authentification détectée
                raise InternalError(_('invalid session'))
            else:
                app_ticket = None
                req = None
                resp = None
                signature_ok = False

                # processing incoming request : LogoutRequest or LogoutResponse (or use different endpoints for each type of request ?)
                if 'SAMLResponse' in request.args:
                    # received a SAML LogoutResponse
                    signature_ok, saml_resp = extract_message(CERTFILE, request, 'SAMLResponse')
                    if not signature_ok:
                        # FIXME TODO: generate a saml negative saml response
                        raise InternalError("%s : %s" % (_('signature error'), str(saml_resp)))
                    response = samlp.LogoutResponseFromString(saml_resp)
                    resp_issuer = response.issuer.text
                    in_response_to = response.in_response_to
                    # checking existence of corresponding request
                    if in_response_to not in self.manager.pending_logout[sso_session][3]:
                        # FIXME TODO: generate a saml negative saml response
                        raise InternalError("%s : %s" % (_('corresponding logout request not found'), in_response_to))
                    issue_instant = date_from_string(response.issue_instant)
                    status_code = response.status.status_code.value
                    if response.status.status_message is not None:
                        status_msg = response.status.status_message.text
                    ticket_id = self.manager.pending_logout[sso_session][3][in_response_to]
                    if status_code == samlp.STATUS_SUCCESS:
                        self.manager.pending_logout[sso_session][1].append(ticket_id)
                    log.msg('SAMLResponse received from %s in response to %s' % (resp_issuer, in_response_to))

                if 'SAMLRequest' in request.args:
                    # received a SAML LogoutRequest
                    signature_ok, saml_req = extract_message(CERTFILE, request, 'SAMLRequest')
                    if not signature_ok:
                        # FIXME TODO: generate a saml negative saml response
                        raise InternalError("signature error : %s" % str(saml_req))
                    saml_req = samlp.LogoutRequestFromString(saml_req)
                    req_issuer = saml_req.issuer.text
                    req_id = saml_req.id
                    app_ticket = saml_req.session_index.text
                    if app_ticket in self.manager.app_sessions:
                        ticket = self.manager.app_sessions[app_ticket]
                        if authenticated and ticket.session_id == sso_session:
                            if req_issuer != ticket.sp_ident:
                                # FIXME TODO: generate a saml negative saml response
                                raise InternalError(_('logout requester (%s) does not own session') % req_issuer)
                            # new logout initiated by a sesison participant
                            self.manager.pending_logout[sso_session] = [samlp.STATUS_SUCCESS, [], (app_ticket, req_id, req_issuer), {}]
                            # we mark this ticket as no longer valid
                            ticket.valid = False
                        else:
                            # FIXME TODO: generate a saml negative saml response
                            raise InternalError(_('logout request for invalid session'))
                    else:
                        # FIXME TODO: generate a saml negative saml response
                        raise InternalError(_('invalid session'))
                    log.msg('SAMLRequest received from %s' % (req_issuer))

                elif sso_session not in self.manager.pending_logout:
                    # new logout initiated by identity provider
                    self.manager.pending_logout[sso_session] = [samlp.STATUS_SUCCESS, [], None, {}]

                # we generate logout requests for all known sessions attached to main SSO session
                # searching saml_sessions not processed previously
                self.manager.pending_logout[sso_session][0] = samlp.STATUS_SUCCESS
                for ticket in self.manager.user_app_tickets[sso_session]:
                    if hasattr(ticket,'sp_ident'):
                        if ticket.valid:
                            sp_identity = ticket.sp_ident
                            # fetching sp metadata
                            sp_meta = self.manager.get_metadata(sp_identity)
                            user_id = ticket.name_id
                            return self.send_logout_request(sp_identity, sp_meta, user_id, ticket, sso_session, from_url)
                        else:
                            # logout already requested for this ticket
                            if ticket.ticket in self.manager.pending_logout[sso_session][1]:
                                closed_services.append(ticket.sp_ident)
                            else:
                                failed_services.append(ticket.sp_ident)
                                # at least one logout request did not return a successfull response
                                self.manager.pending_logout[sso_session][0] = samlp.STATUS_PARTIAL_LOGOUT

                # all participants have been notified, we check logout final status and
                # we send a LogoutResponse matching original request
                if self.manager.pending_logout[sso_session][2]:
                    # fetching logout initiator informations
                    app_ticket, req_id, req_issuer  = self.manager.pending_logout[sso_session][2]
                    # posting response if logout requested by LogoutRequest
                    # fetch service provider url in metadata ?
                    sp_meta = self.manager.get_metadata(req_issuer)
                    service_url, binding, response_url = get_endpoint(sp_meta, 'SingleLogoutService')
                    if response_url:
                        # we use response_url if specified in metadata
                        service_url = response_url
                    # we send final logout response to requester
                    response_id = gen_ticket_id('SAMLResponse', self.manager.address)
                    sign_method, response = saml_response.gen_logout_response(req_id, response_id, IDP_IDENTITY, service_url, CERTFILE, self.manager.pending_logout[sso_session][0])
                    log.msg("sending logout response to %s" % (service_url))
                    # return http.Response(stream=response)
                    if binding == samlp.BINDING_HTTP_REDIRECT:
                        # REDIRECT response
                        response = encode_request(response, True)
                        # la signature doit se faire sur la chaine suivante : SAMLResponse=value&RelayState=value&SigAlg=value -> encodé dans location header
                        req_args = sign_request(CERTFILE, sign_method, 'SAMLResponse', response, relay_state)
                        if req_args != []:
                            redirect_url = '%s?%s' % (service_url, urllib.urlencode(req_args))
                        else:
                            redirect_url = service_url
                        resp = http.RedirectResponse(redirect_url)
                    else:
                        # POST response (default fallback)
                        response = encode_request(response)
                        relay_info = ""
                        if relay_state:
                            relay_info = """<input type="hidden" name="RelayState" value="%s">""" % relay_state
                        form="""
                                <FORM action="%s">
                                <input type="hidden" name="SAMLResponse" value="%s">
                                %s
                                <input type="hidden" name="SigAlg" value="%s">
                                <input type="Submit" value="%s">
                        """ % (relay_state, service_url, response, relay_info, sign_method, _('Proceed to service'))
                        resp = http.Response(stream=gen_page(_('Logout confirmation to service'), form, css))
                else:
                    # message for idp initiated single logout
                    logout_msg = ""
                    if closed_services != []:
                        logout_msg += "<b>%s</b>" % _("Following service providers have been disconnected")
                        logout_msg += "<br/><table border=0><tr><td>%s</td></tr></table>" % "</td></tr><tr><td>".join(closed_services)
                    if failed_services != []:
                        logout_msg += "<b>%s</b>" % _("Following service providers did not report successful logout")
                        logout_msg += "<br/><table border=0><tr><td>%s</td></tr></table>" % "</td></tr><tr><td>".join(failed_services)
                    if from_url:
                        logout_msg += """<form action="%s">
                        <br/><b>%s</b> : <br/><h2><font color="blue">%s</font></h2><br/>
                        <input type="submit" value ="%s"/>
                        </form></p>""" % (from_url, _('Following page has been proposed'), from_url, _('Proceed to above url')) 
                    else:
                        logout_msg += """<br/><form action="/login" method="POST">
                        <input type="Submit" value="%s"></form>""" % _('New session')
                    if self.manager.pending_logout[sso_session][0] == samlp.STATUS_PARTIAL_LOGOUT:
                        logout_status = _("Single Logout partially successfull")
                    else:
                        logout_status = _("Single Logout successful")
                    resp = http.Response(stream=gen_page(logout_status, str(logout_msg), css))

            ############################
            # Logout SSO Session Globale

            # we perform logout for real (close sso session and destroy tickets in cache)
            self.manager.logout(sso_session)

        except InternalError, e:
            err_msg = str(e)

        if from_url and closed_services+failed_services == []:
            resp = http.RedirectResponse(from_url)
            log.msg("logged out : redirecting to : %s" % from_url)

        if err_msg:
            form = format_err(err_msg)
            if from_url:
                form += """<p><FORM action="%s">
                <b>%s</b> : <br/><h2><font color="blue">%s</font></h2><br/>
                <input type="submit" value ="%s"/>
                </form></p>""" % (from_url, _('Following page has been proposed'), from_url, _('Proceed to above url')) 
            else:
                form += """<br/><form action="/login" method="post">
                <input type="Submit" value="%s"></form>""" % _('New session')

            return http.Response(stream=gen_page(_('Logout'), form, css))

        # mise en place des cookies
        if sso_session:
            cookie.expires = 1
            resp.headers.setHeader('Set-Cookie', cookies)
        return resp

class SamlMetadata(SamlResource):

    def __init__(self, manager):
        self.manager = manager
        super(SamlMetadata, self).__init__()

    def _render(self, request):
        """displays metadata for this server"""
        data = gen_metadata(AUTH_FORM_URL, CERTFILE)
        resp = http.Response(stream=data)
        return resp

class SamlErrorPage(SamlResource):
    """
    page d'erreur http
    """
    addSlash = False

    def __init__(self, code, description):
        self.code = code
        self.description = description
        title = ('Eole-SSO : Erreur')
        super(SamlErrorPage, self).__init__()

    def _render(self,request):
        return http.StatusResponse(code=self.code, description=self.description)

class SamlResponse(SamlResource):

    def __init__(self, manager):
        self.manager = manager
        super(SamlResource, self).__init__()

    def remove_old_ticket(self, session_id, sp_ident):
        """ deletes any existing session for a given service provider
        """
        for ticket in self.manager.user_app_tickets[session_id]:
            if getattr(ticket, 'sp_ident', '') == sp_ident:
                old_ticket = ticket.ticket
                # remove this ticket from running logouts if any
                if session_id in self.manager.pending_logout:
                    if old_ticket in self.manager.pending_logout[session_id][1]:
                        self.manager.pending_logout[sso_session][1].remove(old_ticket)
                # remove ticket from user global session
                self.manager.user_app_tickets[session_id].remove(ticket)
                # remove ticket from application tickets
                del(self.manager.app_sessions[old_ticket])
                del(ticket)
                break

    def _render(self, request):
        """Generate a saml2 response using HTTP-post profile

        - verify if SSO session already established
        - if not, establish one
        - create response data and send it to service provider
        """
        # checking request arguments
        if 'SAMLRequest' in request.args:
            # received a SAML AuthnRequest
            saml_request = request.args['SAMLRequest'][0]
            # TODO process AuthnRequest
            # data = process_auth_request(decode_request(saml_request))
        else:
            saml_request = None
        if 'RelayState' in request.args:
            relay_state = request.args['RelayState'][0]
        else:
            relay_state = None
        if 'sp_ident' in request.args:
            sp_ident = request.args['sp_ident'][0]
        else:
            sp_ident = 'sentinelle.eole.lan'
        if 'ticket' in request.args:
            app_ticket = request.args['ticket'][0]
        else:
            app_ticket = None
        try:
            css = request.args['css'][0]
        except KeyError:
            css = DEFAULT_CSS
        # get service provider informations
        sp_meta = self.manager.get_metadata(sp_ident)
        #if 'entityID' in sp_meta:
        #    sp_ident = sp_meta['entityID']
        # check that at least assertion_consumer_url is present
        service_url, binding, response_url = get_endpoint(sp_meta, 'AssertionConsumerService')
        # verifying authentication
        authenticated = False
        session_id = None
        user_id = ""
        user_data = {}
        response_id = gen_random_id('_')
        cookies = getCookies(request)
        from_credentials = False
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                session_id = cookie.value
                # we check this session validity
                authenticated = self.manager.validate_session(session_id)
                if authenticated:
                    # session is simulated through a dummy CAS app_ticket
                    # we use relay_state data to provide filtering on service location
                    # if relay_state is not given or not an url, we use assertion_consumer_url as service
                    if app_ticket:
                        # check if ticket comes from /verify url
                        ticket = self.manager.app_sessions[app_ticket]
                        if ticket.session_id == session_id and \
                           ticket.from_credentials == True and \
                           ticket.service_url == "%s/saml" % (AUTH_FORM_URL):
                            # modifying service in ticket to match relay_state / consumer service_url
                            ticket.service_url = get_service_from_url(relay_state or service_url)
                            ticket.filter = self.manager._check_filter(ticket.service_url)
                        else:
                            app_ticket = None
                    if not app_ticket:
                        # if a session already exists for the service provider, we remove it
                        self.remove_old_ticket(session_id, sp_ident)
                        app_ticket = self.manager.get_app_ticket(session_id, relay_state or service_url)
                        # pour les applications saml, on utilise le filtre saml par défaut (pour avoir au moins l'attribut ctemail)
                        if self.manager.app_sessions[app_ticket].filter == 'default':
                            self.manager.app_sessions[app_ticket].filter = 'saml'
                    if app_ticket:
                        # fetching user_data available for this end service (relay_state)
                        app_ok, user_data = self.manager.get_user_details(app_ticket, relay_state or service_url, renew=False, sections=True, keep_valid=True)
                        data, filter_data = user_data
                        user_data = {}
                        # modification du nom des attributs en fonction du filtre
                        for attrs in filter_data.values():
                            for name, label in attrs.items():
                                user_data[name] = data[label]
                        # determining user id to transmit in response, we use mail_acad attribute
                        # FIXME : random id ?
                        #if NAMEID_FIELD in user_data:
                        #    user_id = user_data[NAMEID_FIELD][0]
                        #else:
                        #    user_id = self.manager.sso_sessions[session_id]
                        user_id = gen_random_id('NAMEID_')
                        # sends a SAML Authentication Response
                        # we store some more info in app_ticket :
                        # sp identifier
                        # response id
                        # authnRequest id if present ?
                        self.manager.app_sessions[app_ticket].sp_ident = sp_ident
                        self.manager.app_sessions[app_ticket].response_id = response_id
                        self.manager.app_sessions[app_ticket].name_id = user_id
                        from_credentials = self.manager.app_sessions[app_ticket].from_credentials


        if authenticated:
            attr_table = []
            for attr, val in user_data.items():
                if type(val) == list:
                    val = ', '.join(val)
                if val == "": val = "&nbsp;"
                attr_table.append('<tr><td>%s</td><td>%s</td></tr>' % (attr, val))
            attr_msg = _('Following attributes will be sent')
            msg = """<table border=1><tr><th colspan=2>%s</th></tr>%s</table>""" % (attr_msg, '\n'.join(attr_table))
        else:
            # invalid authentication
            # msg = """<h2>You are not logged in</h2>"""
            # redirection sur la page de login
            redirect_args = []
            if saml_request:
                redirect_args.append(('SAMLRequest', saml_request))
            if relay_state:
                redirect_args.append(('RelayState', relay_state))
            if sp_ident:
                redirect_args.append(('sp_ident', sp_ident))
            else:
                sp_ident = IDP_IDENTITY
            #req_args = [('redirect_to', "%s/saml?%s" % (AUTH_FORM_URL, urllib.urlencode(redirect_args)))]
            req_args = [('service', "%s/saml?%s" % (AUTH_FORM_URL, urllib.urlencode(redirect_args)))]
            redirect_url = '%s?%s' % (AUTH_FORM_URL, urllib.urlencode(req_args))
            resp = http.RedirectResponse(redirect_url)
            return resp

        sign_method, response = saml_response.gen_response(response_id, authenticated, app_ticket, from_credentials, user_id, user_data, saml_request, service_url, IDP_IDENTITY, sp_ident, AUTH_FORM_URL, CERTFILE)

        if request.args.has_key('show'):
            return http.Response(stream=response)

        if binding == samlp.BINDING_HTTP_REDIRECT:
            # REDIRECT response
            response = encode_request(response, True)
            req_args = sign_request(CERTFILE, sign_method, 'SAMLResponse', response, relay_state)
            if req_args != []:
                redirect_url = '%s?%s' % (service_url, urllib.urlencode(req_args))
            else:
                redirect_url = service_url
            resp = http.RedirectResponse(redirect_url)
        else:
            # POST response (default fallback)
            response = encode_request(response)
            relay_info = ""
            if relay_state:
                relay_info = """<input type="hidden" name="RelayState" value="%s">""" % relay_state
            form="""
                    %s
                    <br/>
                    <FORM action="%s" method="POST">
                    <input type="hidden" name="SAMLResponse" value="%s">
                    %s
                    <input type="hidden" name="SigAlg" value="%s">
                    <input type="Submit" value="%s">
            """ % (msg, service_url, response, relay_info, sign_method, _('Proceed to service'))
            head = "%s : <br/>%s" % (_('You are being Authenticated to service'), relay_state or service_url)
            resp = http.Response(stream=gen_page(head, form, css))

        return resp

    def locateChild(self, request, segments):
        if segments and segments[0] == 'metadata':
            return SamlMetadata(self.manager), ()
        if segments and segments[0] == 'logout':
            return SamlLogout(self.manager), ()
        if segments == ():
            return self, ()
        return SamlErrorPage(responsecode.NOT_FOUND, u"Ressource non trouvée %s" % '/'.join(segments)), ()

